/*************************************************************************
* Rutoken                                                                *
* Copyright (c) 2003-2017, CJSC Aktiv-Soft. All rights reserved.         *
* Подробная информация:  http://www.rutoken.ru                           *
*************************************************************************/

package ru.rutoken.samples.pkcs11Wrapper;

import com.sun.jna.Memory;
import com.sun.jna.Pointer;
import com.sun.jna.NativeLong;
import com.sun.jna.ptr.NativeLongByReference;
import java.util.Random;
import ru.rutoken.Pkcs11.*;

public class Derive {

    private final static NativeLong CKD_CPDIVERSIFY_KDF_JRT = new NativeLong(0x90000009);

    private final static int KEY_VALUE_LENGTH = 64;
    private final static int UKM_LENGTH = 8;

    private final static NativeLongByReference hPubKeyR1 = new NativeLongByReference();
    private final static NativeLongByReference hPrivKeyR1 = new NativeLongByReference();

    private final static NativeLongByReference hPubKeyR2 = new NativeLongByReference();
    private final static NativeLongByReference hPrivKeyR2 = new NativeLongByReference();

    private final static byte[] USER_PIN = {'1', '2', '3', '4', '5', '6', '7', '8'};

    private final static byte[] ATTR_CRYPTO_PRO_A = {(byte)0x06, (byte)0x07, (byte)0x2A, (byte)0x85, (byte)0x03, (byte)0x02, (byte)0x02, (byte)0x23, (byte)0x01};

    private final static CK_ATTRIBUTE[] PublicKeyTmpl1 = (CK_ATTRIBUTE[]) (new CK_ATTRIBUTE()).toArray(7);
    private final static CK_ATTRIBUTE[] PrivateKeyTmpl1 =  (CK_ATTRIBUTE[])(new CK_ATTRIBUTE()).toArray(8);

    private final static CK_ATTRIBUTE[] PublicKeyTmpl2 = (CK_ATTRIBUTE[]) (new CK_ATTRIBUTE()).toArray(7);
    private final static CK_ATTRIBUTE[] PrivateKeyTmpl2 =  (CK_ATTRIBUTE[])(new CK_ATTRIBUTE()).toArray(8);

    private static void initPubKeyTemplate(CK_ATTRIBUTE[] template, String label, String id) {
        template[0].setAttr(Pkcs11Constants.CKA_CLASS, Pkcs11Constants.CKO_PUBLIC_KEY);
        template[1].setAttr(Pkcs11Constants.CKA_LABEL, label);
        template[2].setAttr(Pkcs11Constants.CKA_ID, id);
        template[3].setAttr(Pkcs11Constants.CKA_KEY_TYPE, RtPkcs11Constants.CKK_GOSTR3410);
        template[4].setAttr(Pkcs11Constants.CKA_TOKEN, true);
        template[5].setAttr(Pkcs11Constants.CKA_PRIVATE, false);
        template[6].setAttr(RtPkcs11Constants.CKA_GOSTR3410_PARAMS, ATTR_CRYPTO_PRO_A);
    }

    private static void initPrivKeyTemplate(CK_ATTRIBUTE[] template, String label, String id) {
        template[0].setAttr(Pkcs11Constants.CKA_CLASS, Pkcs11Constants.CKO_PRIVATE_KEY);
        template[1].setAttr(Pkcs11Constants.CKA_LABEL, label);
        template[2].setAttr(Pkcs11Constants.CKA_ID, id);
        template[3].setAttr(Pkcs11Constants.CKA_KEY_TYPE, RtPkcs11Constants.CKK_GOSTR3410);
        template[4].setAttr(Pkcs11Constants.CKA_TOKEN, true);
        template[5].setAttr(Pkcs11Constants.CKA_PRIVATE, true);
        template[6].setAttr(Pkcs11Constants.CKA_DERIVE, true);
        template[7].setAttr(RtPkcs11Constants.CKA_GOSTR3410_PARAMS, ATTR_CRYPTO_PRO_A);
    }

    public static void main(String[] args) {
        NativeLong hSession = Pkcs11Constants.CK_INVALID_HANDLE;
        try {
            System.out.println("Example of C_DeriveKey using");

            System.out.println("Library initialization and acquiring of function list");
            NativeLong rv = RtPkcs11Library.getInstance().C_Initialize(null);
            if (!rv.equals(Pkcs11Constants.CKR_OK))
            {
                throw new Exception("Failed to initialize library, error code:" + Long.toHexString(rv.longValue()));
            }

            System.out.println("Acquiring list of slots with connected tokens");
            NativeLongByReference slotsCount = new NativeLongByReference();
            rv = RtPkcs11Library.getInstance().C_GetSlotList(true, null, slotsCount);
            if (!rv.equals(Pkcs11Constants.CKR_OK))
            {
                throw new Exception("C_GetSlotList failed, error code: " + Long.toHexString(rv.longValue()));
            }
            if (slotsCount.getValue().intValue() == 0)
            {
                throw new Exception("No Rutoken is available!");
            }

            NativeLong[] pSlotList = new NativeLong[slotsCount.getValue().intValue()];
            rv = RtPkcs11Library.getInstance().C_GetSlotList(true, pSlotList, slotsCount);
            if (!rv.equals(Pkcs11Constants.CKR_OK))
            {
                throw new Exception("C_GetSlotList failed, error code: " + Long.toHexString(rv.longValue()));
            }

            System.out.println("Opening session");
            NativeLongByReference phSession = new NativeLongByReference();
            rv = RtPkcs11Library.getInstance().C_OpenSession(pSlotList[0],
                     new NativeLong(Pkcs11Constants.CKF_SERIAL_SESSION.longValue() | Pkcs11Constants.CKF_RW_SESSION.longValue()),
                    null, null, phSession);
            if (!rv.equals(Pkcs11Constants.CKR_OK))
            {
                throw new Exception("C_OpenSession failed, error code: " + Long.toHexString(rv.longValue()));
            }
            hSession = phSession.getValue();

            System.out.println("Logging in as user");
            rv = RtPkcs11Library.getInstance().C_Login(hSession, Pkcs11Constants.CKU_USER, USER_PIN, new NativeLong(USER_PIN.length));
            if (!rv.equals(Pkcs11Constants.CKR_OK))
            {
                throw new Exception("C_Login failed, error code: " + Long.toHexString(rv.longValue()));
            }

            CK_MECHANISM ckmGenMech = new CK_MECHANISM(RtPkcs11Constants.CKM_GOSTR3410_KEY_PAIR_GEN, Pointer.NULL, new NativeLong(0));

            String ckaLabelPubKey1 = "Sample GOST R 34.10-2001 Public Key 1 (Aktiv Co.)";
            String ckaLabelPrivKey1 = "Sample GOST R 34.10-2001 Private Key 1 (Aktiv Co.)";
            String ckaId1 = "Sample GOST R 34.10-2001 Keypair 1 (Aktiv Co.)";

            initPubKeyTemplate(PublicKeyTmpl1, ckaLabelPubKey1, ckaId1);
            initPrivKeyTemplate(PrivateKeyTmpl1, ckaLabelPrivKey1, ckaId1);

            System.out.println("Generating key pair 1");
            rv = RtPkcs11Library.getInstance().C_GenerateKeyPair(hSession, ckmGenMech, PublicKeyTmpl1, new NativeLong(PublicKeyTmpl1.length),
                     PrivateKeyTmpl1, new NativeLong(PrivateKeyTmpl1.length), hPubKeyR1, hPrivKeyR1);
            if (!rv.equals(Pkcs11Constants.CKR_OK))
            {
                throw new Exception("C_GenerateKeyPair failed, error code: " + Long.toHexString(rv.longValue()));
            }

            NativeLong hPubKey1 = hPubKeyR1.getValue();

            String ckaLabelPubKey2 = "Sample GOST R 34.10-2001 Public Key 2 (Aktiv Co.)";
            String ckaLabelPrivKey2 = "Sample GOST R 34.10-2001 Private Key 2 (Aktiv Co.)";
            String ckaId2 = "Sample GOST R 34.10-2001 Keypair 2 (Aktiv Co.)";

            initPubKeyTemplate(PublicKeyTmpl2, ckaLabelPubKey2, ckaId2);
            initPrivKeyTemplate(PrivateKeyTmpl2, ckaLabelPrivKey2, ckaId2);

            System.out.println("Generating key pair 2");
            rv = RtPkcs11Library.getInstance().C_GenerateKeyPair(hSession, ckmGenMech, PublicKeyTmpl2, new NativeLong(PublicKeyTmpl2.length),
                     PrivateKeyTmpl2, new NativeLong(PrivateKeyTmpl2.length), hPubKeyR2, hPrivKeyR2);
            if (!rv.equals(Pkcs11Constants.CKR_OK))
            {
                throw new Exception("C_GenerateKeyPair failed, error code: " + Long.toHexString(rv.longValue()));
            }

            NativeLong hPrivKey2 = hPrivKeyR2.getValue();

            Pointer pValue = new Memory(KEY_VALUE_LENGTH);
            NativeLong pValueLen = new NativeLong(KEY_VALUE_LENGTH);

            CK_ATTRIBUTE[] valueAttr = (CK_ATTRIBUTE[]) (new CK_ATTRIBUTE()).toArray(1);
            valueAttr[0].setAttr(Pkcs11Constants.CKA_VALUE, pValue, pValueLen);

            System.out.println("Getting attribute value");
            rv = RtPkcs11Library.getInstance().C_GetAttributeValue(hSession, hPubKey1, valueAttr, new NativeLong(valueAttr.length));
            if (!rv.equals(Pkcs11Constants.CKR_OK))
            {
                throw new Exception("C_GetAttributeValue failed, error code: " + Long.toHexString(rv.longValue()));
            }

            Random r = new Random();
            byte[] deriveParamUkm = new byte[UKM_LENGTH];
            r.nextBytes(deriveParamUkm);

            CK_GOSTR3410_DERIVE_PARAMS_JRT  deriveParameters = new CK_GOSTR3410_DERIVE_PARAMS_JRT(CKD_CPDIVERSIFY_KDF_JRT, pValueLen,
                     pValue.getByteArray(0, pValueLen.intValue()), new NativeLong(deriveParamUkm.length), deriveParamUkm);

            deriveParameters.write();   //writes the fields of the structure to native memory

            CK_MECHANISM ckmDeriveMech = new CK_MECHANISM(RtPkcs11Constants.CKM_GOSTR3410_DERIVE, deriveParameters.getPointer(), new NativeLong(deriveParameters.size()));

            CK_ATTRIBUTE[] derivedKeyAttr = (CK_ATTRIBUTE[]) (new CK_ATTRIBUTE()).toArray(8);
            derivedKeyAttr[0].setAttr(Pkcs11Constants.CKA_LABEL, "Derived GOST 28147-89 key");
            derivedKeyAttr[1].setAttr(Pkcs11Constants.CKA_CLASS, Pkcs11Constants.CKO_SECRET_KEY);
            derivedKeyAttr[2].setAttr(Pkcs11Constants.CKA_KEY_TYPE, RtPkcs11Constants.CKK_GOST28147);
            derivedKeyAttr[3].setAttr(Pkcs11Constants.CKA_TOKEN, false);
            derivedKeyAttr[4].setAttr(Pkcs11Constants.CKA_MODIFIABLE, true);
            derivedKeyAttr[5].setAttr(Pkcs11Constants.CKA_PRIVATE, true);
            derivedKeyAttr[6].setAttr(Pkcs11Constants.CKA_EXTRACTABLE, true);
            derivedKeyAttr[7].setAttr(Pkcs11Constants.CKA_SENSITIVE, false);
            NativeLong derivedKeyAttrLen = new NativeLong(derivedKeyAttr.length);

            NativeLongByReference phKey = new NativeLongByReference();

            System.out.println("Deriving key");
            rv = RtPkcs11Library.getInstance().C_DeriveKey(hSession, ckmDeriveMech, hPrivKey2, derivedKeyAttr, derivedKeyAttrLen, phKey);
            if (!rv.equals(Pkcs11Constants.CKR_OK))
            {
                throw new Exception("C_DeriveKey failed, error code: " + Long.toHexString(rv.longValue()));
            }
        }
        catch(Exception e){
            System.out.println(e.getMessage());
        }
        finally {
            System.out.println("Deleting public key 1");
            NativeLong rv = RtPkcs11Library.getInstance().C_DestroyObject(hSession, hPubKeyR1.getValue());
            if (!rv.equals(Pkcs11Constants.CKR_OK))
            {
                System.out.println("C_DestroyObject failed, error code: " + Long.toHexString(rv.longValue()));
            }

            System.out.println("Deleting private key 1");
            rv = RtPkcs11Library.getInstance().C_DestroyObject(hSession, hPrivKeyR1.getValue());
            if (!rv.equals(Pkcs11Constants.CKR_OK))
            {
                System.out.println("C_DestroyObject failed, error code: " + Long.toHexString(rv.longValue()));
            }

            System.out.println("Deleting public key 2");
            rv = RtPkcs11Library.getInstance().C_DestroyObject(hSession, hPubKeyR2.getValue());
            if (!rv.equals(Pkcs11Constants.CKR_OK))
            {
                System.out.println("C_DestroyObject failed, error code: " + Long.toHexString(rv.longValue()));
            }

            System.out.println("Deleting private key 2");
            rv = RtPkcs11Library.getInstance().C_DestroyObject(hSession, hPrivKeyR2.getValue());
            if (!rv.equals(Pkcs11Constants.CKR_OK))
            {
                System.out.println("C_DestroyObject failed, error code: " + Long.toHexString(rv.longValue()));
            }

            System.out.println("Logging out");
            rv = RtPkcs11Library.getInstance().C_Logout(hSession);
            if (!rv.equals(Pkcs11Constants.CKR_OK))
            {
                System.out.println("C_Logout failed, error code: " + Long.toHexString(rv.longValue()));
            }

            System.out.println("Closing session");
            rv = RtPkcs11Library.getInstance().C_CloseSession(hSession);
            if (!rv.equals(Pkcs11Constants.CKR_OK))
            {
                System.out.println("C_CloseSession failed, error code: " + Long.toHexString(rv.longValue()));
            }

            System.out.println("Finalizing PKCS11 library");
            rv = RtPkcs11Library.getInstance().C_Finalize(null);
            if (!rv.equals(Pkcs11Constants.CKR_OK))
            {
                System.out.println("C_Finalize failed, error code: " + Long.toHexString(rv.longValue()));
            }

            System.out.println("Test has been completed.");
        }
    }
}